package clientutil

import (
	"context"
	"crypto/tls"
	"errors"
	"log"
	"net"
	"net/http"
	"sync"
	"time"
)

var errAllBackendsFailed = errors.New("all backends failed")

type dnsResolver struct{}

func (r *dnsResolver) ResolveIPs(hosts []string) []string {
	var resolved []string
	for _, hostport := range hosts {
		host, port, err := net.SplitHostPort(hostport)
		if err != nil {
			log.Printf("error parsing %s: %v", hostport, err)
			continue
		}
		hostIPs, err := net.LookupIP(host)
		if err != nil {
			log.Printf("error resolving %s: %v", host, err)
			continue
		}
		for _, ip := range hostIPs {
			resolved = append(resolved, net.JoinHostPort(ip.String(), port))
		}
	}
	return resolved
}

var defaultResolver = &dnsResolver{}

type resolver interface {
	ResolveIPs([]string) []string
}

// Balancer for HTTP connections. It will round-robin across available
// backends, trying to avoid ones that are erroring out, until one
// succeeds or they all fail.
//
// This object should not be used for load balancing of individual
// HTTP requests: once a new connection is established, requests will
// be sent over it until it errors out. It's meant to provide a
// *reliable* connection to a set of equivalent backends for HA
// purposes.
type balancer struct {
	hosts    []string
	resolver resolver
	stop     chan bool

	// List of currently valid (or untested) backends, and ones
	// that errored out at least once.
	mx    sync.Mutex
	addrs []string
	ok    map[string]bool
}

var backendUpdateInterval = 60 * time.Second

// Periodically update the list of available backends.
func (b *balancer) updateProc() {
	tick := time.NewTicker(backendUpdateInterval)
	for {
		select {
		case <-b.stop:
			return
		case <-tick.C:
			resolved := b.resolver.ResolveIPs(b.hosts)
			if len(resolved) > 0 {
				b.mx.Lock()
				b.addrs = resolved
				b.mx.Unlock()
			}
		}
	}
}

// Returns a list of all available backends, split into "good ones"
// (no errors seen since last successful connection) and "bad ones".
func (b *balancer) getBackends() ([]string, []string) {
	b.mx.Lock()
	defer b.mx.Unlock()

	var good, bad []string
	for _, addr := range b.addrs {
		if ok := b.ok[addr]; ok {
			good = append(good, addr)
		} else {
			bad = append(bad, addr)
		}
	}
	return good, bad
}

func (b *balancer) notify(addr string, ok bool) {
	b.mx.Lock()
	b.ok[addr] = ok
	b.mx.Unlock()
}

func netDialContext(ctx context.Context, network, addr string) (net.Conn, error) {
	timeout := 30 * time.Second
	// Go < 1.9 does not have net.DialContext, reimplement it in
	// terms of net.DialTimeout.
	if deadline, ok := ctx.Deadline(); ok {
		timeout = time.Until(deadline)
	}
	return net.DialTimeout(network, addr, timeout)
}

func (b *balancer) dial(ctx context.Context, network, addr string) (net.Conn, error) {
	// Start by attempting a connection on 'good' targets.
	good, bad := b.getBackends()

	for _, addr := range good {
		// Go < 1.9 does not have DialContext, deal with it
		conn, err := netDialContext(ctx, network, addr)
		if err == nil {
			return conn, nil
		} else if err == context.Canceled {
			// A timeout might be bad, set the error bit
			// on the connection.
			b.notify(addr, false)
			return nil, err
		}
		b.notify(addr, false)
	}

	for _, addr := range bad {
		conn, err := netDialContext(ctx, network, addr)
		if err == nil {
			b.notify(addr, true)
			return conn, nil
		} else if err == context.Canceled {
			return nil, err
		}
	}

	return nil, errAllBackendsFailed
}

// NewTransport returns a suitably configured http.RoundTripper that
// talks to a specific backend service. It performs discovery of
// available backends via DNS (using A or AAAA record lookups), tries
// to route traffic away from faulty backends.
//
// It will periodically attempt to rediscover new backends.
func NewTransport(backends []string, tlsConf *tls.Config, resolver resolver) http.RoundTripper {
	if resolver == nil {
		resolver = defaultResolver
	}
	addrs := resolver.ResolveIPs(backends)
	b := &balancer{
		hosts:    backends,
		resolver: resolver,
		addrs:    addrs,
		ok:       make(map[string]bool),
	}
	go b.updateProc()

	return &http.Transport{
		DialContext:     b.dial,
		TLSClientConfig: tlsConf,
	}
}
